@@ -3,11 +3,10 @@ from django.urls import path
 
 from .. import views
 from ..models.image import UploadedImage
-from .base import UploadedFileBase
+from .base import ResourceAdminBase
 
 
-@admin.register(UploadedImage)
-class UploadedImageAdmin(UploadedFileBase):
+class UploadedImageAdminBase(ResourceAdminBase):
     upload_view_class = views.image.UploadFileView
     delete_view_class = views.image.DeleteFileView
     change_view_class = views.image.ChangeFileView
@@ -32,3 +31,8 @@ class UploadedImageAdmin(UploadedFileBase):
             ),
         ]
         return urlpatterns
+
+
+@admin.register(UploadedImage)
+class UploadedImageAdmin(UploadedImageAdminBase):
+    pass
