@@ -1,9 +1,19 @@
+import datetime
+import io
 import os
-from typing import Any, Dict, Iterable, Optional, Tuple
+import pathlib
+import posixpath
+import warnings
+from pathlib import Path
+from typing import Any, Dict, Iterable, Optional, Tuple, Union
 
+from django.core.exceptions import SuspiciousFileOperation
 from django.core.files import File
+from django.core.files.storage import FileSystemStorage, Storage
 from django.db import models
+from django.db.models.base import ModelBase
 from django.db.models.fields.files import FieldFile
+from django.db.models.utils import make_model_tuple
 from django.utils.timezone import now
 from django.utils.translation import gettext_lazy as _
 from PIL import Image
@@ -15,7 +25,28 @@ from ..conf import settings
 from ..files import VariationFile
 from ..typing import FileLike
 from ..variations import PaperVariation
-from .mixins import BacklinkModelMixin, FileFieldProxyMixin, FileProxyMixin
+from .mixins import FileFieldProxyMixin, FileProxyMixin
+
+try:
+    from django.core.files.utils import validate_file_name
+except ImportError:
+    # new in Django 3.1.10
+    def validate_file_name(name, allow_relative_path=False):
+        if os.path.basename(name) in {'', '.', '..'}:
+            raise SuspiciousFileOperation("Could not derive file name from '%s'" % name)
+
+        if allow_relative_path:
+            # Use PurePosixPath() because this branch is checked only in
+            # FileField.generate_filename() where all file paths are expected to be
+            # Unix style (with forward slashes).
+            path = pathlib.PurePosixPath(name)
+            if path.is_absolute() or '..' in path.parts:
+                raise SuspiciousFileOperation(
+                    "Detected path traversal attempt in '%s'" % name
+                )
+        elif name != os.path.basename(name):
+            raise SuspiciousFileOperation("File name '%s' includes path elements" % name)
+        return name
 
 __all__ = [
     "NoPermissionsMetaBase",
@@ -44,7 +75,7 @@ class NoPermissionsMetaBase:
     """
     Отменяет создание автоматических объектов Permission у наследников класса.
     """
-    def __new__(mcs, name, bases, attrs, **kwargs):
+    def __new__(cls, name, bases, attrs, **kwargs):
         meta = attrs.pop("Meta", None)
         if meta is None:
             meta = type("Meta", (), {"default_permissions": ()})
@@ -54,19 +85,63 @@ class NoPermissionsMetaBase:
             meta = type("Meta", meta.__bases__, meta_attrs)
         attrs["Meta"] = meta
 
-        return super().__new__(mcs, name, bases, attrs, **kwargs)
+        return super().__new__(cls, name, bases, attrs, **kwargs)
 
 
 class ResourceBaseMeta(NoPermissionsMetaBase, models.base.ModelBase):
-    pass
+    """
+    Приём, позволяющий переопределить OneToOne-связь между моделями при наследовании
+    от абстрактной модели.
+
+    По умолчанию, при наследовании от абстрактной модели, унаследованной от конкретной
+    (concrete), попытка переопределния OneToOne-связи не замещает поле по умолчанию,
+    а добавляет второе.
+
+    Ссылка:
+    https://docs.djangoproject.com/en/3.2/topics/db/models/#specifying-the-parent-link-field
+    """
+
+    def __new__(mcs, name, bases, attrs, **kwargs):
+        parents = [b for b in bases if isinstance(b, ModelBase)]
+
+        parent_links = {}
+        for base in reversed(parents):
+            # Conceptually equivalent to `if base is Model`.
+            if not hasattr(base, "_meta"):
+                continue
+
+            # Locate OneToOneField instances.
+            for field in base._meta.local_fields:
+                if isinstance(field, models.OneToOneField) and field.remote_field.parent_link:
+                    key = make_model_tuple(field.remote_field.model)
+                    parent_links.setdefault(key, []).append(field)
+
+        for fieldname, field in list(attrs.items()):
+            if isinstance(field, models.OneToOneField) and field.remote_field.parent_link:
+                key = make_model_tuple(field.remote_field.model)
+                if key in parent_links:
+                    inherited_field = parent_links[key].pop()
+                    if fieldname != inherited_field.name:
+                        # force delete inherited field
+                        attrs[inherited_field.name] = None
+
+        return super().__new__(mcs, name, bases, attrs)
 
 
 class ResourceBase(models.Model, metaclass=ResourceBaseMeta):
     """
     Базовый класс ресурса.
     """
-    created_at = models.DateTimeField(_("created at"), default=now, editable=False)
-    modified_at = models.DateTimeField(_("changed at"), auto_now=True, editable=False)
+    created_at = models.DateTimeField(
+        _("created at"),
+        default=now,
+        editable=False
+    )
+    modified_at = models.DateTimeField(
+        _("changed at"),
+        auto_now=True,
+        editable=False
+    )
 
     class Meta:
         abstract = True
@@ -81,16 +156,16 @@ class ResourceBase(models.Model, metaclass=ResourceBaseMeta):
         """
         return {
             "id": self.pk,
-            "created": self.created_at.isoformat(),
-            "modified": self.modified_at.isoformat(),
+            "created": self.created_at.isoformat() if self.created_at else None,
+            "modified": self.modified_at.isoformat() if self.modified_at else None,
         }
 
 
-class Resource(BacklinkModelMixin, ResourceBase):
+class Resource(ResourceBase):
     """
     Ресурс.
-    Включает поля, позволяющие обратиться к модели и полю,
-    через которые был создан ресурс.
+
+    TODO: после переноса миксины BacklinkModelMixin класс остался пустым.
     """
 
     class Meta(ResourceBase.Meta):
@@ -102,8 +177,8 @@ class FileResource(FileProxyMixin, Resource):
     Подкласс ресурса, представляющего файл.
     """
 
-    basename = models.CharField(
-        _("basename"),
+    resource_name = models.CharField(
+        _("resource name"),
         max_length=255,
         editable=False,
         help_text=_("Human-readable resource name"),
@@ -114,60 +189,86 @@ class FileResource(FileProxyMixin, Resource):
         editable=False,
         help_text=_("Lowercase, without leading dot"),
     )
-    size = models.PositiveIntegerField(_("size"), default=0, editable=False)
+    size = models.PositiveIntegerField(
+        _("size"),
+        default=0,
+        editable=False
+    )
     checksum = models.CharField(
         _("checksum"),
         max_length=64,
         editable=False,
     )
-    uploaded_at = models.DateTimeField(_("uploaded at"), default=now, editable=False)
+    uploaded_at = models.DateTimeField(
+        _("uploaded at"),
+        default=now,
+        editable=False
+    )
 
     class Meta(Resource.Meta):
         abstract = True
 
     def __str__(self):
-        return self.get_basename()
+        return self.get_caption()
 
     def __repr__(self):
-        return "{}('{}')".format(type(self).__name__, self.get_basename())
+        return "{}('{}')".format(type(self).__name__, self.name)
+
+    @property
+    def basename(self):
+        warnings.warn(
+            "'basename' is deprecated in favor of 'resource_name'",
+            DeprecationWarning,
+            stacklevel=2
+        )
+        return self.resource_name
+
+    @basename.setter
+    def basename(self, value):
+        warnings.warn(
+            "'basename' is deprecated in favor of 'resource_name'",
+            DeprecationWarning,
+            stacklevel=2
+        )
+        self.resource_name = value
 
     @property
     def name(self) -> str:
         """
         Полное имя загруженного файла.
-        В отличие от `self.basename` это имя включает относительный путь,
-        суффикс и расширение.
+        Это имя включает относительный путь, суффикс и расширение.
         """
         raise NotImplementedError
 
-    def get_basename(self) -> str:
+    def get_caption(self) -> str:
         """
         Человекопонятное имя файла.
         Не содержит суффикса, которое может быть добавлено файловым хранилищем.
         """
         if self.extension:
-            return "{}.{}".format(self.basename, self.extension)
-        return self.basename
+            return "{}.{}".format(self.resource_name, self.extension)
+        return self.resource_name
 
-    def _require_file(self):
+    def get_file_size(self) -> int:
         """
-        Требование наличия непустой ссылки на файл.
-        Физическое существование файла не гарантируется и должно проверяться
-        методом `file_exists`.
+        Получение размера загруженного файла.
         """
-        file = self.get_file()
-        if not file:
-            file_field = self.get_file_field()
-            raise ValueError(_("The '%s' attribute has no file associated with it.") % file_field.name)
+        raise NotImplementedError
+
+    def file_exists(self) -> bool:
+        """
+        Проверка существования файла.
+        """
+        raise NotImplementedError
 
     def as_dict(self) -> Dict[str, Any]:
         return {
             **super().as_dict(),
-            "name": self.basename,
+            "name": self.resource_name,
             "extension": self.extension,
+            "caption": self.get_caption(),
             "size": self.size,
-            "url": self.get_file_url(),
-            "uploaded": self.uploaded_at.isoformat()
+            "uploaded": self.uploaded_at.isoformat() if self.uploaded_at else None
         }
 
     def update_checksum(self, file: FileLike = None) -> bool:
@@ -183,58 +284,37 @@ class FileResource(FileProxyMixin, Resource):
             self.checksum = new_checksum
         return old_checksum != new_checksum
 
-    def get_file(self) -> File:
-        raise NotImplementedError
-
-    def set_file(self, value):
-        """
-        Обновление файла в текущей модели при переименовании / удалении.
-        Применяется в Cloudinary, т.к. файловые storage устанавливают
-        значения поля самостоятельно.
-        """
-        raise NotImplementedError
-
-    def get_file_field(self):
-        """
-        Получение файлового поля модели.
-        """
-        raise NotImplementedError
-
-    def get_file_size(self) -> int:
-        """
-        Получение размера загруженного файла.
-        """
-        raise NotImplementedError
-
-    def get_file_url(self) -> str:
-        """
-        Получение ссылки на загруженный файл.
-        """
-        raise NotImplementedError
-
-    def file_exists(self) -> bool:
-        """
-        Проверка существования файла.
-        """
-        raise NotImplementedError
-
-    def attach_file(self, file: FileLike, name: str = None, **options):
+    def attach(self, file: Union[str, Path, FileLike], name: str = None, **options):
         """
         Присоединение файла к экземпляру ресурса.
-        В действительности, сохранение файла происходит в методе `_attach_file`.
-        Не переопределяйте этот метод, если не уверены в том, что вы делаете.
+        Этот метод является обёрткой. Фактическое сохранение файла
+        происходит в методе `_attach()`.
 
         Если на данном этапе обнаруживается, что переданный файл не может
         быть представлен этой моделью, необходимо вызвать исключение
-        UnsupportedFileError.
+        UnsupportedResource.
         """
-        if not isinstance(file, File):
-            name = name or getattr(file, "name", None)
-            if name:
-                name = os.path.basename(name)
-            file = File(file, name=name)
-        elif name:
+        if not name:
+            if isinstance(file, str):
+                name = file
+            elif isinstance(file, Path):
+                name = str(file)
+            else:
+                name = getattr(file, "name", None)
+
+        # prevent path traversal
+        if name:
+            name = os.path.basename(name)
+
+        # convert to `django.core.files.File` instance
+        need_to_close = False
+        if isinstance(file, (str, Path)):
+            need_to_close = True
+            file = File(open(file, "rb"), name=name)
+        elif isinstance(file, File):
             file.name = name
+        else:
+            file = File(file, name=name)
 
         prepared_file = self._prepare_file(file, **options)
 
@@ -246,7 +326,7 @@ class FileResource(FileProxyMixin, Resource):
         if prepared_file.seekable():
             prepared_file.seek(0)
 
-        response = self._attach_file(prepared_file, **options)
+        attach_result = self._attach(prepared_file, **options)
 
         self.size = self.get_file_size()
         self.uploaded_at = now()
@@ -266,28 +346,48 @@ class FileResource(FileProxyMixin, Resource):
             instance=self,
             file=self.get_file(),
             options=options,
-            response=response,
+            response=attach_result,
         )
 
+        if need_to_close:
+            if not file.closed:
+                file.close()
+
+    def attach_file(self, *args, **kwargs):
+        warnings.warn(
+            "attach_file() is deprecated in favor of attach()",
+            DeprecationWarning,
+            stacklevel=2
+        )
+        return self.attach(*args, **kwargs)
+
     def _prepare_file(self, file: File, **options) -> File:
         """
         Подготовка файла к присоединению к модели.
         """
         return file
 
-    def _attach_file(self, file: File, **options):
+    def _attach(self, file: File, **options):
         raise NotImplementedError
 
-    def rename_file(self, new_name: str, **options):
+    def _attach_file(self, *args, **kwargs):
+        warnings.warn(
+            "_attach_file() is deprecated in favor of _attach()",
+            DeprecationWarning,
+            stacklevel=2
+        )
+        return self._attach(*args, **kwargs)
+
+    def rename(self, new_name: str, **options):
         """
         Переименование файла.
-        В действительности, переименование файла происходит в методе `_rename_file`.
-        Не переопределяйте этот метод, если не уверены в том, что вы делаете.
+        Этот метод является обёрткой. Фактическое переименование файла
+        происходит в методе `_rename()`.
         """
         self._require_file()
 
         if not self.file_exists():
-            raise exceptions.FileNotFoundError(self)
+            raise FileNotFoundError(self.name)
 
         old_name = self.name
 
@@ -299,7 +399,7 @@ class FileResource(FileProxyMixin, Resource):
             options=options,
         )
 
-        response = self._rename_file(new_name, **options)
+        response = self._rename(new_name, **options)
 
         self.modified_at = now()
 
@@ -312,14 +412,30 @@ class FileResource(FileProxyMixin, Resource):
             response=response,
         )
 
-    def _rename_file(self, new_name: str, **options):
+    def rename_file(self, *args, **kwargs):
+        warnings.warn(
+            "rename_file() is deprecated in favor of rename()",
+            DeprecationWarning,
+            stacklevel=2
+        )
+        return self.rename(*args, **kwargs)
+
+    def _rename(self, new_name: str, **options):
         raise NotImplementedError
 
+    def _rename_file(self, *args, **kwargs):
+        warnings.warn(
+            "_rename_file() is deprecated in favor of _rename()",
+            DeprecationWarning,
+            stacklevel=2
+        )
+        return self._rename(*args, **kwargs)
+
     def delete_file(self, **options):
         """
         Удаление файла.
-        В действительности, удаление файла происходит в методе `_delete_file`.
-        Не переопределяйте этот метод, если не уверены в том, что вы делаете.
+        Этот метод является обёрткой. Фактическое удаление файла
+        происходит в методе `_delete_file()`.
         """
         signals.pre_delete_file.send(sender=type(self), instance=self, options=options)
         response = self._delete_file(**options)
@@ -337,28 +453,79 @@ class FileFieldResource(FileFieldProxyMixin, FileResource):
     class Meta(FileResource.Meta):
         abstract = True
 
+    def _require_file(self):
+        file = self.get_file()
+        if not file:
+            file_field = self.get_file_field()
+            raise ValueError(_("The '%s' attribute has no file associated with it.") % file_field.name)
+
+    def _attach(self, file: File, **options):
+        self.get_file().save(file.name, file, save=False)
+        self.resource_name = helpers.get_filename(file.name)
+        self.extension = helpers.get_extension(self.name)
+
+    def _rename(self, new_name: str, **options):
+        # копирование файла, чтобы не вредить кэшированию
+        file = self.get_file()
+        with file.open() as fp:
+            file.save(new_name, fp, save=False)
+
+        self.resource_name = helpers.get_filename(new_name)
+        self.extension = helpers.get_extension(self.name)
+
+    def _delete_file(self, **options):
+        self.get_file().delete(save=False)
+
     @property
     def name(self) -> str:
         self._require_file()
         return self.get_file().name
 
+    def get_file_field(self):
+        """
+        Получение файлового поля модели.
+        """
+        raise NotImplementedError
+
     def get_file_folder(self) -> str:
         """
         Возвращает путь к папке, в которую будет сохранен файл.
-        Результат вызова используется в параметре `upload_to` в случае Django storage.
         """
-        return ""
+        raise NotImplementedError
+
+    def get_file_storage(self) -> Storage:
+        raise NotImplementedError
+
+    def generate_filename(self, filename: str) -> str:
+        """
+        Формирование каталога и имени файла при сохранении.
+        """
+        dirname = self.get_file_folder()
+        dirname = datetime.datetime.now().strftime(dirname)
+        filename = posixpath.join(dirname, filename)
+        filename = validate_file_name(filename, allow_relative_path=True)
+
+        storage = self.get_file_storage()
+        return storage.generate_filename(filename)
 
     def get_file(self) -> FieldFile:
         raise NotImplementedError
 
+    def set_file(self, value):
+        field = self.get_file_field()
+        setattr(self, field.attname, value)
+
     def get_file_size(self) -> int:
         self._require_file()
         return self.get_file().size
 
     def get_file_url(self) -> str:
-        self._require_file()
-        return self.get_file().url
+        warnings.warn(
+            "get_file_url() is deprecated in favor of 'url' property",
+            DeprecationWarning,
+            stacklevel=2
+        )
+        return self.url
 
     def file_exists(self) -> bool:
         file = self.get_file()
@@ -366,22 +533,11 @@ class FileFieldResource(FileFieldProxyMixin, FileResource):
             return False
         return file.storage.exists(file.name)
 
-    def _attach_file(self, file: File, **options):
-        self.get_file().save(file.name, file, save=False)
-
-        self.basename = helpers.get_filename(file.name)
-        self.extension = helpers.get_extension(self.name)
-
-    def _rename_file(self, new_name: str, **options):
-        file = self.get_file()
-        with file.open() as fp:
-            file.save(new_name, fp, save=False)
-
-        self.basename = helpers.get_filename(new_name)
-        self.extension = helpers.get_extension(self.name)
-
-    def _delete_file(self, **options):
-        self.get_file().delete(save=False)
+    def as_dict(self) -> Dict[str, Any]:
+        return {
+            **super().as_dict(),
+            "url": self.url,
+        }
 
 
 class ImageFileResourceMixin(models.Model):
@@ -427,7 +583,7 @@ class ImageFileResourceMixin(models.Model):
         try:
             image = Image.open(file)
         except OSError:
-            raise exceptions.UnsupportedFileError(
+            raise exceptions.UnsupportedResource(
                 _("File `%s` is not an image") % file.name
             )
         else:
@@ -493,27 +649,26 @@ class VersatileImageResourceMixin(ImageFileResourceMixin):
             else:
                 self.recut()
 
-    def attach_file(self, file: FileLike, name: str = None, **options):
-        super().attach_file(file, name=name, **options)  # noqa: F821
+    def attach(self, file: FileLike, name: str = None, **options):
+        super().attach(file, name=name, **options)  # noqa: F821
         self.need_recut = True
         self._setup_variation_files()
 
-    def _delete_file(self):
-        self.delete_variations()
-        super()._delete_file()  # noqa: F821
-
-    def _rename_file(self, new_name: str, **options):
-        super()._rename_file(new_name)  # noqa: F821
-        self.recut()
+    def rename(self, new_name: str, **options):
+        super().rename(new_name, **options)  # noqa: F821
+        self.need_recut = True
         self._setup_variation_files()
 
+    def delete_file(self, **options):
+        super().delete_file(**options)  # noqa: F821
+        self.delete_variations()
+
     def get_variations(self) -> Dict[str, PaperVariation]:
         raise NotImplementedError
 
     def delete_variations(self):
         for vname, vfile in self.variation_files():
             vfile.delete()
-        self._reset_variation_files()
 
     def variation_files(self) -> Iterable[Tuple[str, VariationFile]]:
         if not self._variation_files_cache:
@@ -547,8 +702,17 @@ class VersatileImageResourceMixin(ImageFileResourceMixin):
         Запись изображения в файловое хранилище
         """
         variation_file = self.get_variation_file(name)
-        with variation_file.open("wb") as fp:
-            variation.save(image, fp)
+
+        if isinstance(variation_file.storage, FileSystemStorage):
+            with variation_file.open("wb") as fp:
+                variation.save(image, fp)
+        else:
+            # Не все Storage-классы позволяют записывать контент с помощью вызовов
+            # `open()` и `write()`. Для них нужно использовать метод `_save()`.
+            with io.BytesIO() as buffer:
+                variation.save(image, buffer)
+                content = File(buffer, name=variation_file.name)
+                variation_file.storage._save(variation_file.name, content)
 
     def recut(self, names: Iterable[str] = ()):
         """
@@ -556,7 +720,7 @@ class VersatileImageResourceMixin(ImageFileResourceMixin):
         Можно указать имена конкретных вариаций в параметре `names`.
         """
         if not self.file_exists():
-            raise exceptions.FileNotFoundError(self)
+            raise FileNotFoundError(self.name)
 
         file = self.get_file()
         with file.open() as source:
@@ -571,8 +735,12 @@ class VersatileImageResourceMixin(ImageFileResourceMixin):
                 image = variation.process(img)
                 self._save_variation(name, variation, image)
 
+                vfile = self.get_variation_file(name)
+                self.__dict__[name] = vfile
+                self._variation_files_cache[name] = vfile
+
                 signals.variation_created.send(
-                    sender=type(self), instance=self, file=self.get_variation_file(name)
+                    sender=type(self), instance=self, file=vfile
                 )
 
     def recut_async(self, names: Iterable[str] = ()):
